# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import os
import datetime

from mock import Mock, patch
from fixture import TempIO

from rattail.db import model
from rattail.db.changes import record_changes
from rattail.tests import RattailTestCase

from rattail_locsms import changes


class ChangesProcessorTestCase(RattailTestCase):

    def setUp(self):
        self.setup_rattail()
        self.processor = changes.ChangesProcessor(self.config)


class TestCommonMethods(ChangesProcessorTestCase):

    def test_utc_timestamp_with_no_date(self):
        result = self.processor.utc_timestamp(None, datetime.time(18, 1))
        self.assertTrue(result is None)

    def test_utc_timestamp_with_date_and_no_time(self):
        self.config.set('rattail', 'timezone.rattail.sw.locsms', 'America/Los_Angeles')
        result = self.processor.utc_timestamp(datetime.date(2014, 7, 26), None)
        self.assertEqual(result, datetime.datetime(2014, 7, 26, 7, 0))

    def test_utc_timestamp_with_date_and_time(self):
        self.config.set('rattail', 'timezone.rattail.sw.locsms', 'America/Los_Angeles')
        result = self.processor.utc_timestamp(datetime.date(2014, 7, 26), datetime.time(18, 3))
        self.assertEqual(result, datetime.datetime(2014, 7, 27, 1, 3))

    def test_utcnow(self):
        before = datetime.datetime.utcnow()
        result = self.processor.utcnow()
        after = datetime.datetime.utcnow()
        self.assertTrue(before <= result)
        self.assertTrue(result <= after)

    def test_price_is_current_with_start_and_end(self):
        price = model.ProductPrice(
            starts=datetime.datetime(2014, 7, 1),
            ends=datetime.datetime(2014, 7, 31))
        self.processor.utcnow = Mock()
        self.processor.utcnow.return_value = datetime.datetime(2014, 6, 30)
        self.assertFalse(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 1)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 31)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 8, 1)
        self.assertFalse(self.processor.price_is_current(price))

    def test_price_is_current_with_start_but_no_end(self):
        price = model.ProductPrice(
            starts=datetime.datetime(2014, 7, 1))
        self.processor.utcnow = Mock()
        self.processor.utcnow.return_value = datetime.datetime(2014, 6, 30)
        self.assertFalse(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 1)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 31)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 8, 1)
        self.assertTrue(self.processor.price_is_current(price))

    def test_price_is_current_with_end_but_no_start(self):
        price = model.ProductPrice(
            ends=datetime.datetime(2014, 7, 31))
        self.processor.utcnow = Mock()
        self.processor.utcnow.return_value = datetime.datetime(2014, 6, 30)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 1)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 7, 31)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(2014, 8, 1)
        self.assertFalse(self.processor.price_is_current(price))

    def test_price_is_current_with_no_start_and_no_end(self):
        price = model.ProductPrice()
        self.processor.utcnow = Mock()
        self.processor.utcnow.return_value = datetime.datetime(1900, 1, 1)
        self.assertTrue(self.processor.price_is_current(price))
        self.processor.utcnow.return_value = datetime.datetime(3000, 1, 1)
        self.assertTrue(self.processor.price_is_current(price))


class TestProcessCltChg(ChangesProcessorTestCase):

    def process(self, **kwargs):
        row = dict(
            F1148=u'',          # Customer Id
            F1149=u'',          # Cust. first name
            F1150=u'',          # Cust. last name
            F1154=u'',          # Cust. group
            F1155=u'',          # Cust. company name
            F1172=u'',          # Cust. phone number
            F1173=u'',          # Cust. fax number
            F1573=u'',          # Email address
            F1777=u'',          # Customer group 2
            )
        row.update(**kwargs)
        return self.processor.process_clt_chg(row, self.session)

    def test_empty_customer_is_created(self):
        self.assertEqual(self.session.query(model.Customer).count(), 0)
        customer = self.process(F1148=u'42')
        self.assertEqual(self.session.query(model.Customer).count(), 1)
        self.assertTrue(self.session.query(model.Customer).one() is customer)
        self.assertEqual(len(customer.phones), 0)
        self.assertEqual(len(customer.emails), 0)
        self.assertEqual(len(customer.groups), 0)

    def test_new_customer_receives_correct_attributes(self):
        customer = self.process(
            F1148=u'42',
            F1149=u'Fred',
            F1150=u'Flintstone',
            F1154=u'G1',
            F1155=u'Fred Flintstone',
            F1172=u'417.555.1234',
            F1173=u'417.555.5678',
            F1573=u'fred@mailinator.com',
            F1777=u'G2')
        self.assertEqual(customer.id, u'42')
        self.assertEqual(customer.name, u'Fred Flintstone')
        self.assertEqual(customer.person.first_name, u'Fred')
        self.assertEqual(customer.person.last_name, u'Flintstone')
        self.assertEqual(customer.phone.number, u'417.555.1234')
        self.assertEqual(customer.phones[1].number, u'417.555.5678')
        self.assertEqual(customer.email.address, u'fred@mailinator.com')
        self.assertEqual(customer.groups[0].id, u'G1')
        self.assertEqual(customer.groups[1].id, u'G2')

    def test_attributes_are_overwritten_for_existing_customer(self):
        person = model.Person(first_name=u'Barney', last_name=u'Rubble')
        customer = model.Customer(id=u'42', name=u'Barney Rubble')
        customer.people.append(person)
        customer.add_phone_number('417.555.9876', 'Voice')
        customer.add_phone_number('417.555.5432', 'Voice')
        customer.add_email_address(u'barney@mailinator.com')
        customer.groups.append(model.CustomerGroup(id=u'G3'))
        customer.groups.append(model.CustomerGroup(id=u'G4'))
        self.session.add(customer)
        self.session.commit()

        customer = self.process(
            F1148=u'42',        # matches above
            F1149=u'Fred',
            F1150=u'Flintstone',
            F1154=u'G1',
            F1155=u'Fred Flintstone',
            F1172=u'417.555.1234',
            F1173=u'417.555.5678',
            F1573=u'fred@mailinator.com',
            F1777=u'G2')

        self.assertEqual(customer.id, u'42')
        self.assertEqual(customer.name, u'Fred Flintstone')

        # person name is updated
        self.assertEqual(len(customer.people), 1)
        self.assertEqual(customer.person.first_name, 'Fred')
        self.assertEqual(customer.person.last_name, 'Flintstone')

        # phones are completely replaced
        self.assertEqual(len(customer.phones), 2)
        self.assertEqual(customer.phones[0].number, u'417.555.1234')
        self.assertEqual(customer.phones[1].number, u'417.555.5678')

        # emails are completely replaced
        self.assertEqual(len(customer.emails), 1)
        self.assertEqual(customer.email.address, u'fred@mailinator.com')

        # groups are completely replaced
        self.assertEqual(len(customer.groups), 2)
        self.assertEqual(customer.groups[0].id, u'G1')
        self.assertEqual(customer.groups[1].id, u'G2')

    def fred(self):
        person = model.Person(first_name=u'Fred', last_name=u'Flintstone')
        customer = model.Customer(id=u'42', name=u'Fred Flintstone')
        customer.people.append(person)
        self.session.add(customer)
        return customer

    def test_customer_with_fax_only_gets_phone_inserted_before_fax(self):
        customer = self.fred()
        customer.add_phone_number('417.555.5678', 'Voice')
        self.session.commit()

        customer = self.process(
            F1148='42',
            F1172='417.555.1234',
            F1173='417.555.5678')

        self.assertEqual(len(customer.phones), 2)
        self.assertEqual(customer.phones[0].number, '417.555.1234')
        self.assertEqual(customer.phones[0].type, 'Voice')
        self.assertEqual(customer.phones[1].number, '417.555.5678')
        self.assertEqual(customer.phones[1].type, 'Voice')

    def test_customer_phone_number_may_be_replaced_with_fax(self):
        customer = self.fred()
        customer.add_phone_number('417.555.1234', 'Voice')
        self.session.commit()

        customer = self.process(
            F1148='42',
            F1173='417.555.5678')

        self.assertEqual(len(customer.phones), 1)
        self.assertEqual(customer.phones[0].number, '417.555.5678')
        self.assertEqual(customer.phones[0].type, 'Voice')
        
    def test_customer_fax_number_may_be_overwritten(self):
        customer = self.fred()
        customer.add_phone_number('417.555.1234', 'Voice')
        self.session.commit()

        customer = self.process(
            F1148='42',
            F1173='417.555.5678')

        self.assertEqual(len(customer.phones), 1)
        self.assertEqual(customer.phones[0].number, '417.555.5678')
        self.assertEqual(customer.phones[0].type, 'Voice')

    def test_customer_fax_number_may_be_deleted(self):
        customer = self.fred()
        customer.add_phone_number('417.555.5678', 'Voice')
        self.session.commit()

        customer = self.process(
            F1148=u'42')

        self.assertEqual(len(customer.phones), 0)
        
    def test_customer_email_address_may_be_deleted(self):
        customer = self.fred()
        customer.add_email_address(u'fred@mailinator.com')
        self.session.commit()

        customer = self.process(
            F1148=u'42')

        self.assertEqual(len(customer.emails), 0)
        
    def test_customer_group1_may_be_overwritten(self):
        customer = self.fred()
        customer.groups.append(model.CustomerGroup(id=u'G1'))
        self.session.commit()

        customer = self.process(
            F1148=u'42',
            F1777=u'G2')

        self.assertEqual(len(customer.groups), 1)
        self.assertEqual(customer.groups[0].id, u'G2')
        
    def test_customer_with_no_groups_may_be_assigned_to_group2_only(self):
        customer = self.fred()
        self.session.commit()

        customer = self.process(
            F1148=u'42',
            F1777=u'G2')

        self.assertEqual(len(customer.groups), 1)
        self.assertEqual(customer.groups[0].id, u'G2')

    def test_customer_with_two_groups_may_be_assigned_to_group1_only(self):
        customer = self.fred()
        customer.groups.append(model.CustomerGroup(id=u'G2'))
        customer.groups.append(model.CustomerGroup(id=u'G3'))
        self.session.commit()

        customer = self.process(
            F1148=u'42',
            F1154=u'G1')

        self.assertEqual(len(customer.groups), 1)
        self.assertEqual(customer.groups[0].id, u'G1')

    def test_customer_with_two_groups_may_be_assigned_to_group2_only(self):
        customer = self.fred()
        customer.groups.append(model.CustomerGroup(id=u'G1'))
        customer.groups.append(model.CustomerGroup(id=u'G3'))
        self.session.commit()

        customer = self.process(
            F1148=u'42',
            F1777=u'G2')

        self.assertEqual(len(customer.groups), 1)
        self.assertEqual(customer.groups[0].id, u'G2')

    def test_customer_with_two_groups_may_be_assigned_to_none(self):
        customer = self.fred()
        customer.groups.append(model.CustomerGroup(id=u'G1'))
        customer.groups.append(model.CustomerGroup(id=u'G2'))
        self.session.commit()

        customer = self.process(
            F1148=u'42')

        self.assertEqual(len(customer.groups), 0)


class TestProcessCltChgWithRecord(TestProcessCltChg):

    def setUp(self):
        super(TestProcessCltChgWithRecord, self).setUp()
        record_changes(self.session)


class TestProcessCltDel(ChangesProcessorTestCase):

    def process(self, **kwargs):
        row = dict(
            F1148=u'',          # Customer Id
            )
        row.update(**kwargs)
        self.processor.process_clt_del(row, self.session)

    def test_nonexistent_customer_is_ignored(self):
        customer = model.Customer(id=u'42')
        self.session.add(customer)
        self.session.commit()
        self.assertEqual(self.session.query(model.Customer).count(), 1)
        self.process(F1148=u'99')
        self.assertEqual(self.session.query(model.Customer).count(), 1)

    def test_existent_customer_is_deleted(self):
        customer = model.Customer(id=u'42')
        self.session.add(customer)
        self.session.commit()
        self.assertEqual(self.session.query(model.Customer).count(), 1)
        self.process(F1148=u'42')
        self.assertEqual(self.session.query(model.Customer).count(), 0)


class TestProcessCltDelWithRecord(TestProcessCltDel):

    def setUp(self):
        super(TestProcessCltDelWithRecord, self).setUp()
        record_changes(self.session)


# Fake custom processor, used to test ProcessChanges.
CustomProcessor = Mock()


class TestProcessChanges(RattailTestCase):

    def setUp(self):
        self.setup_rattail()
        self.processor = changes.ProcessChanges(self.config)
        self.tmp = TempIO()

    def test_file_which_does_not_match_pattern_is_ignored(self):
        path = self.tmp.putfile('file-with-funky-name', '')
        with patch('rattail_locsms.changes.ChangesProcessor') as ChangesProcessor:
            self.processor(path)
            self.assertFalse(ChangesProcessor.called)
        self.assertTrue(os.path.exists(path))

    def test_file_which_matches_pattern_but_is_not_end_file_is_ignored(self):
        path = self.tmp.putfile('00000001_OBJ_CHG.CSV', '')
        with patch('rattail_locsms.changes.ChangesProcessor') as ChangesProcessor:
            self.processor(path)
            self.assertFalse(ChangesProcessor.called)
        self.assertTrue(os.path.exists(path))

    def test_end_file_triggers_processing(self):
        begin_path = self.tmp.putfile('00000001_CHANGES_BEGIN', '')
        obj_chg_path = self.tmp.putfile('00000001_OBJ_CHG.CSV', '')
        end_path = self.tmp.putfile('00000001_CHANGES_END', '')
        with patch('rattail_locsms.changes.ChangesProcessor') as ChangesProcessor:
            self.processor(end_path)
            ChangesProcessor.assert_called_once_with(self.config)
            # TODO: Make this really test something.
            self.assertTrue(ChangesProcessor.return_value.process_all_changes.called)
            # assert_called_once_with({'OBJ_CHG': obj_chg_path})
        self.assertFalse(os.path.exists(begin_path))
        self.assertFalse(os.path.exists(obj_chg_path))
        self.assertFalse(os.path.exists(end_path))
        self.assertTrue(os.path.exists(self.tmp.join('processed')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_CHANGES_BEGIN')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_OBJ_CHG.CSV')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_CHANGES_END')))

    def test_custom_processor_factory(self):
        begin_path = self.tmp.putfile('00000001_CHANGES_BEGIN', '')
        obj_chg_path = self.tmp.putfile('00000001_OBJ_CHG.CSV', '')
        end_path = self.tmp.putfile('00000001_CHANGES_END', '')
        self.processor(end_path, processor_factory='rattail_locsms.tests.test_changes:CustomProcessor')
        CustomProcessor.assert_called_once_with(self.config)
        # TODO: Make this really test something.
        self.assertTrue(CustomProcessor.return_value.process_all_changes.called)
        # assert_called_once_with({'OBJ_CHG': obj_chg_path})
        self.assertFalse(os.path.exists(begin_path))
        self.assertFalse(os.path.exists(obj_chg_path))
        self.assertFalse(os.path.exists(end_path))
        self.assertTrue(os.path.exists(self.tmp.join('processed')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_CHANGES_BEGIN')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_OBJ_CHG.CSV')))
        self.assertTrue(os.path.exists(self.tmp.join('processed', '00000001_CHANGES_END')))

    def test_custom_backup_dir(self):
        begin_path = self.tmp.putfile('00000001_CHANGES_BEGIN', '')
        obj_chg_path = self.tmp.putfile('00000001_OBJ_CHG.CSV', '')
        end_path = self.tmp.putfile('00000001_CHANGES_END', '')
        backup_dir = self.tmp.join('custom-backup-dir')
        # backup dir will be created during processing
        self.assertFalse(os.path.exists(backup_dir))
        with patch('rattail_locsms.changes.ChangesProcessor') as ChangesProcessor:
            self.processor(end_path, backup_dir=backup_dir)
            ChangesProcessor.assert_called_once_with(self.config)
            # TODO: Make this really test something.
            self.assertTrue(ChangesProcessor.return_value.process_all_changes.called)
            # assert_called_once_with({'OBJ_CHG': obj_chg_path})
        self.assertFalse(os.path.exists(begin_path))
        self.assertFalse(os.path.exists(obj_chg_path))
        self.assertFalse(os.path.exists(end_path))
        self.assertTrue(os.path.exists(self.tmp.join('custom-backup-dir')))
        self.assertTrue(os.path.exists(self.tmp.join('custom-backup-dir', '00000001_CHANGES_BEGIN')))
        self.assertTrue(os.path.exists(self.tmp.join('custom-backup-dir', '00000001_OBJ_CHG.CSV')))
        self.assertTrue(os.path.exists(self.tmp.join('custom-backup-dir', '00000001_CHANGES_END')))
