# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

from mock import patch
from fixture import TempIO

from sqlalchemy import create_engine

from locsms.db import Session as SmsSession
from locsms.db import model as locsms

from rattail_locsms import transfers


class TestProcessTransfer(TestCase):

    def setUp(self):
        sms_engine = create_engine('sqlite://')
        locsms.Product.__table__.create(sms_engine)
        locsms.Cost.__table__.create(sms_engine)
        SmsSession.configure(bind=sms_engine)
        self.sms_session = SmsSession()
        self.tmp = TempIO()
        self.outbox = self.tmp.mkdir('outbox')

    def tearDown(self):
        self.sms_session.close()
        SmsSession.configure(bind=None)

    def make_transfer(self, filename=None, contents=None):
        if filename is None:
            filename = 'T001901-00000001-0000-002.EXP'
        if contents is None:
            contents = """\
F01
007430500132
"""
        return self.tmp.putfile(filename, contents)

    def test_no_costs_deployed_if_transfer_file_is_empty(self):
        path = self.make_transfer(contents='')
        with patch('rattail_locsms.transfers.deploy_costs') as deploy_costs:
            transfers.process_transfer(path, self.outbox)
            self.assertFalse(deploy_costs.called)

    def test_no_costs_deployed_if_product_not_found(self):
        path = self.make_transfer()
        with patch('rattail_locsms.transfers.deploy_costs') as deploy_costs:
            transfers.process_transfer(path, self.outbox)
            self.assertFalse(deploy_costs.called)

    def test_no_costs_deployed_if_product_has_no_authz_cost(self):
        self.sms_session.add(locsms.Product(F01='007430500132'))
        self.sms_session.commit()
        path = self.make_transfer()
        with patch('rattail_locsms.transfers.deploy_costs') as deploy_costs:
            transfers.process_transfer(path, self.outbox)
            self.assertFalse(deploy_costs.called)

    def test_cost_add_batch_is_deployed_if_product_has_authz_cost_but_no_cost_for_selling_store(self):
        product = locsms.Product(F01='007430500132')
        product.costs.append(locsms.Cost(F1000='PAL', F27='3788', F1184='CASE'))
        self.sms_session.add(product)
        self.sms_session.commit()
        path = self.make_transfer()
        with patch('rattail_locsms.transfers.deploy_costs') as deploy_costs:
            transfers.process_transfer(path, self.outbox)
            self.assertTrue(deploy_costs.called)
